from datetime import date, datetime

from havc.services.file import File
from havc.utils.bytes_conversion import set_converted_bytes_with_label


class Output:
    """
    output and statistics?
    """
    original_files_size = 0.0
    reduced_files_size = 0.0
    files_not_encoded = []
    number_of_files_encoded = 0
    skipped_files = []
    video_files = []
    output_file = None

    def __add_file_information(self, timestamp, video_file, successful):
        self.__add_line('[' + str(timestamp) + ']\nThe following video file was converted: ' + video_file.name_only)

        if not successful:
            self.__add_line('\t*** This file was not encoded successfully! ***\n')
            self.files_not_encoded.append(video_file.absolute_path + video_file.extension)
            return

        self.__add_line('\t-Converted from ' + video_file.extension + ' extension to ' + video_file.target_extension)
        self.__set_message_with_size('\t-The original file size was: ', video_file.original_size)
        self.__set_message_with_size('\t-The converted file size is: ', video_file.converted_size)

        self.__add_line('\n')
        self.number_of_files_encoded += 1
        self.original_files_size += video_file.original_size
        self.reduced_files_size += video_file.converted_size

    def add_file(self, video_file, successful):
        self.video_files.append({
            'timestamp': datetime.utcnow(),
            'video_file': video_file,
            'successful': successful
        })

    def add_skipped_file(self, skipped_file_path):
        self.skipped_files.append(skipped_file_path)

    def process(self, absolute_path_parent):
        date_now = '[' + str(date.today().year) + '-' + str(date.today().month) + '-' + str(date.today().day) + ' ' + str(
            datetime.utcnow().hour) + '-' + str(
            datetime.utcnow().minute) + '-' + str(datetime.utcnow().second) + ']'
        self.output_file = File(absolute_path_parent + '\\output' + date_now + '.txt')
        self.output_file.open('a')

        for video_file in self.video_files:
            self.__add_file_information(video_file['timestamp'], video_file['video_file'], video_file['successful'])

        self.__add_line('[' + str(datetime.utcnow()) + ']\nFinal Statistics:')
        self.__set_message_with_size('\tSize of all original video files: ', self.original_files_size)
        self.__set_message_with_size('\tSize of all converted video files: ', self.reduced_files_size)

        space_saved = self.original_files_size - self.reduced_files_size
        if space_saved > 0:
            self.__set_message_with_size('\tSpace in disk saved: ', space_saved)

        number_of_files_not_encoded = len(self.files_not_encoded)
        self.__add_line('\n\t- Total number of files converted: ' + str(self.number_of_files_encoded))
        self.__add_line('\t- Total number of files unsuccessfully converted: ' + str(number_of_files_not_encoded))
        total = self.number_of_files_encoded + number_of_files_not_encoded
        self.__add_line('\t- Total number of files found: ' + str(total))

        if number_of_files_not_encoded != 0:
            self.__add_line('\n\tThe following video files were not encoded successfully:')
            for file in self.files_not_encoded:
                self.__add_line('\t--> ' + file)

        if len(self.skipped_files) > 0:
            self.__add_line('\n\tThe following video files were skipped:')
            for file in self.skipped_files:
                self.__add_line('\t--> ' + file)

        self.output_file.close()

    def __set_message_with_size(self, message, size):
        size_with_label = set_converted_bytes_with_label(size)
        self.__add_line(message + '%.2f' % size_with_label['size'] + ' ' + size_with_label['label'])

    def __add_line(self, message):
        self.output_file.write(message + '\n')



import os.path
import shutil
from havc.services.directory import Directory


class VideoFile:
    extension = ''
    absolute_path = ''
    name_only = ''
    original_size = 0
    converted_size = 0

    def __init__(self, name, root, file_extensions_to_convert, target_extension):
        self.__name = name
        self.__root = root
        self.__file_extensions_to_convert = file_extensions_to_convert
        self.target_extension = target_extension

    def process(self):
        """
        first checks if the file is to convert, and if not, returns False immediately without processing the rest.
        processes file to store the file name, the extension and its absolute path.
        :return: True if the file processes completely. False if the file is not to convert.
        """
        if not self.__process_extension_file():
            return False

        self.__process_file_name()
        self.__process_absolute_path()

        return True

    def __process_file_sizes_before_after(self):
        self.original_size = os.path.getsize(self.absolute_path + self.extension)

        new_file = Directory(self.absolute_path + self.target_extension)

        self.converted_size = os.path.getsize(new_file.root)

    def copy_to(self, new_path):
        """
        copy original video file to a folder to be deleted later. removes said file from original location.
        :param new_path: new location for the video file
        """
        self.__process_file_sizes_before_after()

        directory = Directory(new_path)
        original = r'{}'.format(self.absolute_path) + self.extension
        target = directory.create(self.name_only + self.extension)

        shutil.copyfile(original, target)

        directory = Directory(self.absolute_path + self.extension)
        directory.remove()

    def already_exists(self):
        return Directory(self.absolute_path + self.target_extension).exists()

    def __process_file_name(self):
        file_name = self.__name.split('.')
        file_name.pop()
        self.name_only = '.'.join(file_name)

    def __process_extension_file(self):
        file_name_extension = '.' + self.__name.split('.')[len(self.__name.split('.')) - 1]
        for ext in self.__file_extensions_to_convert:
            if ext == file_name_extension:
                self.extension = ext
                return True
        return False

    def __process_absolute_path(self):
        directory = Directory(self.__root)
        self.absolute_path = directory.create(self.name_only)





class Argument:
    """
    This class is responsible for creating a new object containing important information about an argument
    """
    def __init__(self, name, abbreviation_name, full_name, help_message, metavar, default='None'):
        """
        initialization of important information about an argument
        :param name: the basic name of the argument
        :param abbreviation_name: the abbreviation of the name to use in the command line
        :param full_name: the full name to use in the command line
        """
        self.name = name
        self.abbreviation_name = abbreviation_name
        self.full_name = full_name
        self.value = None
        self.help_message = help_message
        self.metavar = metavar
        self.default = default

    def set_argument_value(self, value):
        """
        sets the value for the current argument
        :param value: value the argument has
        """
        self.value = value


class Extensions(Argument):

    def set_argument_value(self, value):
        if not isinstance(value, list):
            if not value.startswith('.'):
                value = '.' + value
        else:
            for ext in value:
                if not ext.startswith('.'):
                    value[value.index(ext)] = '.' + ext

        self.value = value



from havc.services.directory import Directory
from havc.utils.error import throw
from havc.entities.prog_arguments import ProgArguments
from havc.configurations.configurations import Configurations
from havc.utils.progsettings import get_version
import argparse


class Arguments:
    """
    This class is responsible for configuring all arguments required for the program to work including saving all mandatory argument values in a
    configuration file for easy frequent use. This way the user only has to configure the first time it runs the program or if the configuration
    file does not exist.
    """

    def __init__(self):
        self.args = None
        self.original_arguments = None
        self.prog_arguments = ProgArguments()
        self.are_configs_saved = False

    def configure(self):
        """
        create and configure arguments to save in a configuration file
        :return: returns all arguments either from the command line or saved configuration file
        """

        self.args = argparse.ArgumentParser(prog="Handbrake Automatic Video Converter",
                                            usage="an automatic video converter/encoder using the HandBrake CLI",
                                            description="this program is meant to convert/encode video from various formats using the rules and "
                                                        "syntax of "
                                                        "the HandBrake CLI. the following commands should explain the arguments.")

        config_file = Configurations()

        self.are_configs_saved = config_file.is_configured()

        self.__add_arguments()

        self.original_arguments = self.args.parse_args()

        self.__check_any_errors()

        config_file.set_original_arguments(self.original_arguments)

        return config_file.process()

    def __add_arguments(self):
        """
        configures and adds the arguments required for the program
        """
        self.args.add_argument('--version', action='version', version='%(prog)s ' + get_version())

        self.args.add_argument(self.prog_arguments.root.abbreviation_name, self.prog_arguments.root.full_name,
                               required=not self.are_configs_saved,
                               help=self.prog_arguments.root.help_message,
                               metavar=self.prog_arguments.root.metavar,
                               default=argparse.SUPPRESS)

        self.args.add_argument(self.prog_arguments.folder_path_to_convert.abbreviation_name,
                               self.prog_arguments.folder_path_to_convert.full_name,
                               required=not self.are_configs_saved,
                               help=self.prog_arguments.folder_path_to_convert.help_message,
                               metavar=self.prog_arguments.folder_path_to_convert.metavar,
                               default=argparse.SUPPRESS)

        self.args.add_argument(self.prog_arguments.original_extensions.abbreviation_name, self.prog_arguments.original_extensions.full_name,
                               action='extend', nargs='+',
                               required=not self.are_configs_saved,
                               help=self.prog_arguments.original_extensions.help_message,
                               metavar=self.prog_arguments.original_extensions.metavar,
                               default=argparse.SUPPRESS)

        self.args.add_argument(self.prog_arguments.target_extension.abbreviation_name, self.prog_arguments.target_extension.full_name,
                               required=not self.are_configs_saved,
                               help=self.prog_arguments.target_extension.help_message,
                               metavar=self.prog_arguments.target_extension.metavar,
                               default=argparse.SUPPRESS)

        self.args.add_argument(self.prog_arguments.deleted_folder.abbreviation_name, self.prog_arguments.deleted_folder.full_name,
                               required=False,
                               help=self.prog_arguments.deleted_folder.help_message,
                               default=self.prog_arguments.deleted_folder.default,
                               metavar=self.prog_arguments.deleted_folder.metavar)

        self.args.add_argument(self.prog_arguments.custom_command.abbreviation_name, self.prog_arguments.custom_command.full_name,
                               required=False,
                               help=self.prog_arguments.custom_command.help_message,
                               default=self.prog_arguments.custom_command.default,
                               metavar=self.prog_arguments.custom_command.metavar)

        self.args.add_argument(self.prog_arguments.safety_question.full_name,
                               action=argparse.BooleanOptionalAction,
                               required=False,
                               help=self.prog_arguments.safety_question.help_message,
                               default=self.prog_arguments.safety_question.default,
                               metavar=self.prog_arguments.safety_question.metavar)

        self.args.add_argument(self.prog_arguments.shutdown_when_done.full_name,
                               action=argparse.BooleanOptionalAction,
                               required=False,
                               help=self.prog_arguments.shutdown_when_done.help_message,
                               default=self.prog_arguments.shutdown_when_done.default,
                               metavar=self.prog_arguments.shutdown_when_done.metavar)

    def __check_any_errors(self):
        if self.__target_and_original_extensions_are_the_same():
            throw('target extension cannot be the same as any of the original file extensions.')

        try:
            if not self.__given_argument_path_exists(self.original_arguments.root):
                throw(self.original_arguments.root + '\' path does not exist.')

            if not self.__given_argument_path_exists(self.original_arguments.convert):
                throw(self.original_arguments.convert + '\' path does not exist.')
        except (AttributeError, TypeError):
            pass

    def __target_and_original_extensions_are_the_same(self):
        try:
            for ext in self.original_arguments.extensions:
                if ext in self.original_arguments.target:
                    return True
            return False
        except (AttributeError, TypeError):
            return False

    @staticmethod
    def __given_argument_path_exists(path):
        argument_path = Directory(path)
        return argument_path.exists()



from havc.entities.video_file import VideoFile
from havc.output import Output
from havc.services.directory import Directory
from havc.command import Command


class Search:
    def __init__(self, arguments):
        self.folder_path = arguments.folder_path_to_convert.value
        self.root_path = arguments.root.value
        self.original_file_extensions = arguments.original_extensions.value
        self.target_file_extension = arguments.target_extension.value
        self.delete_folder = arguments.deleted_folder.value
        self.custom_command = arguments.custom_command.value

    def search(self):
        main_directory = Directory(self.folder_path)
        output_file = Output()
        found_files = False

        delete_folder = self.create_delete_folder()

        for root, dirs, files in main_directory.search_through():
            if self.delete_folder in root:
                continue

            for video in files:
                current_video_file = VideoFile(video, root, self.original_file_extensions, self.target_file_extension)

                if not current_video_file.process():
                    continue

                if current_video_file.already_exists():
                    print(root + '\\' + video + ' was skipped because file with same name already exists!\n\n')
                    output_file.add_skipped_file(root + '\\' + video)
                    continue

                found_files = True
                handbrake = Command(self.root_path)

                if self.custom_command.upper() != 'off'.upper():
                    handbrake.set_custom_command(self.custom_command)

                successful = handbrake.run_command(current_video_file)

                if successful:
                    print('\nEncoding successfully done!\n\n')
                    sub_delete_folder = self.create_delete_sub_folder(root, delete_folder)
                    current_video_file.copy_to(sub_delete_folder)
                else:
                    print('\nEncoding unsuccessful.\n\n')

                output_file.add_file(current_video_file, successful)

        if found_files:
            output_file.process(main_directory.root)
            print('\nAn output file with a summary was created in the following directory: \n\t\t' + output_file.output_file.path + '')
        else:
            print('\nNo files were found with current extensions.')

    def create_delete_folder(self):
        root_directory = Directory(self.folder_path)
        new_delete_directory = Directory(root_directory.last_folder_path)
        to_delete_folder_path = new_delete_directory.create_folder(self.delete_folder)
        return Directory(to_delete_folder_path)

    def create_delete_sub_folder(self, root, delete_folder):
        delete_folder_path = delete_folder.root
        main_directory = Directory(self.folder_path)

        path_name_list = root.split('\\')
        pass_main_folder = False
        for folder in path_name_list:
            if folder == main_directory.current_folder:
                pass_main_folder = True
                
            if pass_main_folder and folder not in delete_folder_path:
                delete_folder_path = delete_folder_path + '\\' + folder

        return delete_folder.create_folder(delete_folder_path)



import textwrap

from havc.entities.argument import Argument, Extensions


class ProgArguments:
    """
    This class is responsible for creating an object containing all arguments used in the program
    """

    def __init__(self):
        self.root = Argument('root',
                             '-r',
                             '--root',
                             'absolute path to the following file: HandBrakeCLI.exe. example: --root \'C:\\path\\to\\folder\'',
                             "")

        self.folder_path_to_convert = Argument('convert',
                                               '-c',
                                               '--convert',
                                               'absolute path to the folder with convertible videos. example: --convert \'C:\\path\\to\\folder\'',
                                               "")

        self.original_extensions = Extensions('extensions',
                                              '-e',
                                              '--extensions',
                                              'list of video\'s extensions to find and convert (with or without \'.\'). example: --extensions .mp4 '
                                              'm4v',
                                              "")

        self.target_extension = Extensions('target',
                                           '-t',
                                           '--target',
                                           'a target video extension to apply when a video is converted',
                                           "")

        self.deleted_folder = Argument('delete_folder',
                                       '-d',
                                       '--delete-folder',
                                       'name of the folder containing original files. default is: \'TO-DELETE\'',
                                       "",
                                       default='TO-DELETE')

        self.custom_command = Argument('custom_command',
                                       '-cc',
                                       '--custom-command',
                                       textwrap.dedent('''\
                                       a custom command inserted by the user. placeholders must be used for the original file and the converted 
                                       file. this command is supposed to work dynamically for a list of files searched in a folder, so:\n'
                                       'example: "--preset "Very Fast 1080p30" -i {of} -o {cf}"\n'
                                       '{of} => original file\n'
                                       '{cf} => converted file\n'
                                       'strings in the command must be between single quotes ->\'\'<-\n'
                                       'if the user wants to no longer use the custom command saved in the configs, the following can be inputted 
                                       instead: -cc OFF'''),
                                       "",
                                       default='OFF')

        self.safety_question = Argument('safety_question',
                                        '',
                                        '--safety-question',
                                        '[WARNING: make sure the folder to convert path is the correct one] enable/disable safety question '
                                        'regarding modifying files in path of folder to convert: (default is enabled)'
                                        'True: --safety-question | False: --no-safety-question',
                                        "",
                                        default=True)

        self.shutdown_when_done = Argument('shutdown',
                                           '',
                                           '--shutdown',
                                           'enable/disable if computer will be shutdown when the program is done: (default is disabled)'
                                           'True: --shutdown | False: --no-shutdown',
                                           "",
                                           default=False)

    def to_list(self):
        arguments = [
            self.root,
            self.folder_path_to_convert,
            self.original_extensions,
            self.target_extension,
            self.deleted_folder,
            self.custom_command,
            self.safety_question,
            self.shutdown_when_done
        ]
        return arguments

    def from_list(self, arguments):
        self.root = arguments[0]
        self.folder_path_to_convert = arguments[1]
        self.original_extensions = arguments[2]
        self.target_extension = arguments[3]
        self.deleted_folder = arguments[4]
        self.custom_command = arguments[5]
        self.safety_question = arguments[6]
        self.shutdown_when_done = arguments[7]



import os

import yaml

from havc.entities.prog_arguments import ProgArguments
from havc.services.directory import Directory
from havc.services.file import File


class Configurations:
    original_arguments = None
    arguments = ProgArguments()
    settings = {}
    path = os.path.dirname(os.path.realpath(__file__)) + '\\userconfigs.yaml'
    file = File(path)

    def set_original_arguments(self, original_arguments):
        self.original_arguments = original_arguments

    def process(self):
        if self.is_configured():
            configs_file = self.file.open('r')
            self.settings = yaml.safe_load(configs_file)
            self.file.close()

        self.__configure()

        return self.arguments

    def is_configured(self):
        directory = Directory(self.file.path)
        if not directory.exists():
            return False

        if self.file.is_empty():
            return False
        return True

    def __configure(self):
        """
        creates a new file and writes all mandatory arguments
        """
        arguments = self.arguments.to_list()

        for configuration in arguments:
            self.settings[configuration.name] = self.__process_configuration(configuration)

        configs_file = self.file.open('w')
        yaml.safe_dump(self.settings, configs_file)
        self.file.close()

        self.arguments.from_list(arguments)

    def __process_configuration(self, configuration):
        if configuration.name not in self.original_arguments:
            if not self.is_configured():
                argument_value = configuration.default
            else:
                try:
                    argument_value = self.settings[configuration.name]
                except KeyError:
                    argument_value = configuration.default
        else:
            argument_value = getattr(self.original_arguments, configuration.name)

        configuration.set_argument_value(argument_value)
        return configuration.value



def split_string(string, separator, stop_character, include_separator=False):
    """
    Splits a given string into a list using a separator except when a sub-string is between a specific stop character.
    :param include_separator: boolean to specify if the separator character must be included in the list or not
    :param string: string to be split
    :param separator: separator to split the string
    :param stop_character: split does not happen between two of these characters
    :return: a list containing the string but separated between specified separator
    """
    list_of_params = []
    param = ''

    is_between = False
    for character in string:
        if is_between:
            if character == stop_character:
                is_between = False
                if not include_separator:
                    continue
        else:
            if character == stop_character:
                is_between = True
                if not include_separator:
                    continue

            if character == separator:
                list_of_params.append(param)
                param = ''
                continue

        param += character

    list_of_params.append(param)

    return list_of_params



import subprocess

from havc.services.directory import Directory
from havc.utils.error import throw
from havc.utils.split_string import split_string


class Command:
    """
    here it will go the future feature about choosing a default command versus a command inserted by the user
    """
    handbrake_command = ''
    custom_command = None
    original_file_placeholder = '{of}'
    converted_file_placeholder = '{cf}'

    def __init__(self, root_path):
        self.root_path = root_path
        self.handbrake_executable = self.root_path + '\\HandBrakeCLI.exe'

    def set_custom_command(self, custom_command):
        self.custom_command = custom_command

    def run_command(self, file_info):
        executable_path = Directory(self.handbrake_executable)
        if not executable_path.exists():
            throw(executable_path.root + ' path is not valid. either the path is incorrect or HandBrakeCLI.exe file does not exist in this path.')

        if self.custom_command is None:
            process = self.__run_static_command(file_info)
        else:
            process = self.__run_custom_command(file_info)

        return self.__success_of(process)

    def __run_static_command(self, file_info):
        basic_command = [
            self.handbrake_executable,
            '--preset', 'Very Fast 1080p30',
            '-i',
            file_info.absolute_path + file_info.extension,
            '-o',
            file_info.absolute_path + file_info.target_extension
        ]

        return self.__run(basic_command)

    def __run_custom_command(self, file_info):
        list_of_args = split_string(self.custom_command, ' ', '\'')

        list_of_args[list_of_args.index(self.original_file_placeholder)] = file_info.absolute_path + file_info.extension
        list_of_args[list_of_args.index(self.converted_file_placeholder)] = file_info.absolute_path + file_info.target_extension

        list_of_args.insert(0, self.handbrake_executable)

        return self.__run(list_of_args)

    @staticmethod
    def __success_of(process):
        result = 'Encode done!' in str(process)
        return result

    @staticmethod
    def __run(args):
        print('Current Command: ' + ' '.join([str(elem) for elem in args]) + '\n')
        return subprocess.run(args, stderr=subprocess.PIPE)



import subprocess
import sys

from havc.arguments import Arguments
from havc.search import Search


def main():
    arguments = Arguments().configure()

    print('All files found with the a ', end="")
    first = True
    for extension in arguments.original_extensions.value:
        if first:
            print(extension, end="")
            first = False
            continue
        print(' or ' + extension, end="")
    print(
        ' extension are going to be converted to ' + arguments.target_extension.value + ' extension and be kept in the "' + arguments.deleted_folder.value + '" folder.')
    print('"' + arguments.folder_path_to_convert.value + '" is going to be modified permanently.\n')

    if arguments.shutdown_when_done.value:
        print('The computer will be shutdown when this program is done.\n')

    videos = Search(arguments)

    if arguments.safety_question.value:
        response = input(
            'Are you sure you want to continue? \n [Y/n]')
        if response == 'n' or response == 'N':
            sys.exit()
        elif response == 'Y':
            videos.search()
            return

    videos.search()

    if arguments.shutdown_when_done.value:
        subprocess.run(["shutdown", "-s"])


if __name__ == '__main__':
    main()




def set_converted_bytes_with_label(size_in_bytes):
    kilobyte = 1024
    megabyte = kilobyte * kilobyte
    gigabyte = megabyte * kilobyte

    if kilobyte < size_in_bytes < megabyte:
        converted_size = size_in_bytes / kilobyte
        label = 'KB'
    elif megabyte < size_in_bytes < gigabyte:
        converted_size = size_in_bytes / megabyte
        label = 'MB'
    elif size_in_bytes >= gigabyte:
        converted_size = size_in_bytes / gigabyte
        label = 'GB'
    else:
        converted_size = size_in_bytes
        label = 'B'

    return {'size': converted_size, 'label': label}
